/* Set up a dialog for the calculation of the visual horizon from
   a given height, and vice versa.
 */

#include <stdio.h>
#include <stdlib.h>
#if !defined(_Windows) && !defined (__MSDOS__)
  #include <sys/types.h>
  #include <unistd.h>
#endif
#include <string.h>
#include <errno.h>
#include <math.h>

#include "../cgitools/cgitools.h"

#define CHUNK 10

struct progdata_s
{
   FILE *errfp;
   char lang[2+1];
   char page[4+1];
   char *scriptname;
} progdata;

enum algor_enum
{
   ALGOR_h, /* h is input, calculate the other ones */
   ALGOR_d, /* d is input, calculate the other ones */
   ALGOR_D, /* D is input, calculate the other ones */
};

struct formdata_s
{
   char h_string[20];
   int  height_in_km;
   char d_string[20];
   char D_string[20];
   char E_string[20];
   enum algor_enum Algor;
} formdata;

void get_invocation_parms (struct progdata_s *p_progdata);

void showform (
   struct progdata_s *p_progdata,
   struct formdata_s *p_formdata);

void getparms (int len, struct formdata_s *p_formdata);

int calculate (struct formdata_s *p_formdata);

static char ErrorMessageBuffer[512];
static void ErrorOnScreen (void)
{
   printf("Content-type: text/html\n\n");
   printf("<html>\n");
   printf("<p>%s</p>\n", ErrorMessageBuffer);
   printf("</html>\n");

   fflush(stdout);
}

int main (int argc, char **argv)
{
   int len;
   char *ErrFile = "vizhoriz/vizhoriz.err";
   char *ErrFileMode = "at";

   progdata.scriptname = "../cgi-bin/vizhoriz.cgi";

   progdata.errfp = fopen(ErrFile, ErrFileMode);
   if (!progdata.errfp)
   {
       strncpy(ErrorMessageBuffer, strerror(errno), sizeof ErrorMessageBuffer - 1);
       #if !defined(_Windows) && !defined (__MSDOS__)
          sprintf(ErrorMessageBuffer + strlen(ErrorMessageBuffer),
             "<br>Cannot open file %s in mode \"%s\", <br>user/grp = %d/%d, eff. user/grp = %d/%d\n",
             ErrFile, ErrFileMode, getuid(), geteuid(), getgid(), getegid());
       #else
          sprintf(ErrorMessageBuffer + strlen(ErrorMessageBuffer),
             "<br>Cannot open file %s in mode \"%s\"\n",
             ErrFile, ErrFileMode);
       #endif
       ErrorOnScreen();
       return 1;
   }

   get_invocation_parms(&progdata);

   if ((len = userinput()) == 0)
   {
      /* Initialise values */
      strcpy(formdata.h_string, "385");
      strcpy(formdata.E_string, "6365");
      /* Calculate other values */
      formdata.Algor = ALGOR_h;
      calculate(&formdata);

      showform(&progdata, &formdata);
   }
   else
   {
      getparms(len, &formdata);
      calculate(&formdata);
      showform(&progdata, &formdata);
   }

   fclose(progdata.errfp);

   return 0;
}

void get_invocation_parms (struct progdata_s *p_progdata)
{
   char *query;
   char **Vars = NULL, **V;

   strcpy(p_progdata->lang, "en");
   strcpy(p_progdata->page, "text");

   query = getenv("QUERY_STRING");
   if (query)
   {
      Vars = CgiConvVars(query, NULL);

      for (V = Vars; V[0]; V += 2)
      {
         if (strcmp(V[0], "lang") == 0 && strlen(V[1]))
         {
            sprintf(p_progdata->lang, "%2.2s", V[1]);
         }
         else if (strcmp(V[0], "page") == 0 && strlen(V[1]))
         {
            sprintf(p_progdata->page, "%4.4s", V[1]);
         }
      }
   }
}

#define CHUNK 10

void showform (
   struct progdata_s *p_progdata,
   struct formdata_s *p_formdata)
{
   char htmlfile[41];
   FILE *fp;
   int bufsize = CHUNK;
   char *buf;
   char *b;
   char c;

   sprintf(htmlfile, "vizhoriz/%.2s/viz-%.4s.htm",
      p_progdata->lang, p_progdata->page);

   fp = fopen(htmlfile, "rt");
   if (!fp)
   {
      time_t now;
      fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
      fprintf(p_progdata->errfp, "Cannot open file %s, errno=%d = %s\n",
         htmlfile, errno, strerror(errno));
      return;
   }

   buf = malloc(bufsize);
   if (!buf)
   {
      time_t now;
      fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
      fprintf(p_progdata->errfp, "Cannot malloc %d bytes\n", bufsize);
      return;
   }
   b = buf;

   while ((c = getc(fp)) != EOF)
   {
      if (b - buf >= bufsize)
      {
         buf = realloc(buf, bufsize + CHUNK);
         if (!buf)
         {
            time_t now;
            fprintf(p_progdata->errfp, "%s", ctime(((now = time(NULL)), &now)));
            fprintf(p_progdata->errfp,
               "Cannot malloc %d bytes\n", bufsize + CHUNK);
            return;
         }
         b = buf + bufsize;
         bufsize += CHUNK;
      }
      *b++ = (char)c;
   }
   *b = '\0';

   fclose(fp);

   printf("Content-type: text/html\n\n");
   printf(buf, p_progdata->scriptname, p_progdata->lang,
         p_formdata->h_string,
        !p_formdata->height_in_km ? "Checked" : "",
         p_formdata->height_in_km ? "Checked" : "",
         p_formdata->d_string,
         p_formdata->D_string,
         p_formdata->E_string);
   fflush(stdout);
}

void getparms (int len, struct formdata_s *p_formdata)
{
   char **Vars = NULL, **V;
   char *buf = calloc(len + 1, 1);

   if (!buf ||
      fread(buf, len, 1, stdin) != 1)
   {
      free(buf);
      return;
   }

   Vars = CgiConvVars(buf, NULL);

   for (V = Vars; V[0]; V += 2)
   {
      if (strcmp(V[0], "Val_h") == 0 && strlen(V[1]))
      {
         sprintf(p_formdata->h_string, "%-15.15s", V[1]);
      }
      else if (strcmp(V[0], "Val_d") == 0)
      {
         sprintf(p_formdata->d_string, "%-15.15s", V[1]);
      }
      else if (strcmp(V[0], "Val_DD") == 0)
      {
         sprintf(p_formdata->D_string, "%-15.15s", V[1]);
      }
      else if (strcmp(V[0], "Val_E") == 0)
      {
         sprintf(p_formdata->E_string, "%-15.15s", V[1]);
      }
      else if (strcmp(V[0], "Calc_h") == 0)
      {
         p_formdata->Algor = ALGOR_h;
      }
      else if (strcmp(V[0], "Calc_d") == 0)
      {
         p_formdata->Algor = ALGOR_d;
      }
      else if (strcmp(V[0], "Calc_DD") == 0)
      {
         p_formdata->Algor = ALGOR_D;
      }
      else if (strcmp(V[0], "mkm") == 0)
      {
         p_formdata->height_in_km = strcmp(V[1], "m");
      }
   }

   free(buf);
   if (Vars)
      free(Vars);
}

int calculate (struct formdata_s *p_formdata)
{
   double h, h_m, h_km, HorizonDoorLucht, HorizonLangsAarde, EarthRadius;

   EarthRadius = atof(p_formdata->E_string);

   switch (p_formdata->Algor)
   {
      case ALGOR_h:
         h = atof(p_formdata->h_string);
         if (h <= 0.0)
            h = 1.0;

         h_m  = p_formdata->height_in_km ? h * 1000.0 : h;
         h_km = p_formdata->height_in_km ? h : h / 1000.0;
         HorizonDoorLucht  = sqrt(h_km * (h_km + 2 * EarthRadius));
         HorizonLangsAarde = EarthRadius *
            acos(EarthRadius / (EarthRadius + h_km));
         break;

      case ALGOR_d:
         HorizonDoorLucht = atof(p_formdata->d_string);
         if (HorizonDoorLucht <= 0.0)
            HorizonDoorLucht = 1.0;

         h_km = sqrt(
               EarthRadius * EarthRadius
               +
               HorizonDoorLucht * HorizonDoorLucht
               )
            - EarthRadius;
         h_m = 1000 * h_km,
         HorizonLangsAarde = EarthRadius *
            acos(EarthRadius / (EarthRadius + h_km));
         break;

      case ALGOR_D:
         HorizonLangsAarde  = atof(p_formdata->D_string);
         if (HorizonLangsAarde <= 0.0)
            HorizonLangsAarde = 1.0;
         if (HorizonLangsAarde >= EarthRadius * M_PI_2)
            HorizonLangsAarde = EarthRadius * M_PI_2 - 0.001;

         HorizonDoorLucht = EarthRadius *
            tan(HorizonLangsAarde / EarthRadius);
         h_km = sqrt(
               EarthRadius * EarthRadius
               +
               HorizonDoorLucht * HorizonDoorLucht
               )
            - EarthRadius;
         h_m = 1000 * h_km;
         break;
   }
   sprintf(p_formdata->h_string, "%7.3f",
      p_formdata->height_in_km ? h_km : h_m);
   sprintf(p_formdata->d_string, "%7.3f", HorizonDoorLucht);
   sprintf(p_formdata->D_string, "%7.3f", HorizonLangsAarde);
   sprintf(p_formdata->E_string, "%7.0f", EarthRadius);

   return 0;
}
